#[cfg(test)]
mod test;

use std::{collections::HashMap, rc::Rc};

use crate::{Ast, Constant, Ident, Type};

impl Ast {
    pub fn beta_reduce(self) -> Ast {
        match self {
            Ast::Application(lhs, rhs) => match *lhs {
                Ast::Abstraction(var, _, ast) => ast.subst(var, *rhs).beta_reduce(),
                lhs => Ast::Application(Box::new(lhs), rhs),
            },
            t => t,
        }
    }

    fn subst(self, var: Ident, subst: Ast) -> Ast {
        match self {
            Ast::Abstraction(var1, typ, ast) => {
                if var != var1 {
                    Ast::Abstraction(var1, typ, Box::new(ast.subst(var, subst)))
                } else {
                    Ast::Abstraction(var1, typ, ast)
                }
            }
            Ast::Application(lhs, rhs) => Ast::Application(
                Box::new(lhs.subst(var.clone(), subst.clone())),
                Box::new(rhs.subst(var, subst)),
            ),
            Ast::Variable(v) if v == var => subst,
            Ast::Variable(v) => Ast::Variable(v),
            Ast::Constant(constant) => Ast::Constant(constant),
        }
    }
}

impl Ast {
    pub fn to_de_brujin(self) -> DeBrujinAst {
        self.to_de_brujin_inter(Rc::new(HashMap::new()))
    }

    fn to_de_brujin_inter(self, mut gamma: Rc<HashMap<String, usize>>) -> DeBrujinAst {
        match self {
            Ast::Abstraction(var, _, ast) => {
                let gamma_ref = Rc::make_mut(&mut gamma);
                gamma_ref.values_mut().for_each(|v| *v += 1);
                gamma_ref.insert(var, 1);

                DeBrujinAst::Abstraction(Box::new(ast.to_de_brujin_inter(gamma)))
            }
            Ast::Application(lhs, rhs) => DeBrujinAst::Application(
                Box::new(lhs.to_de_brujin_inter(gamma.clone())),
                Box::new(rhs.to_de_brujin_inter(gamma)),
            ),
            Ast::Variable(v) => {
                if let Some(c) = gamma.get(&v) {
                    DeBrujinAst::BoundVariable(*c)
                } else {
                    DeBrujinAst::FreeVariable(v)
                }
            }
            Ast::Constant(constant) => DeBrujinAst::Constant(constant),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub enum DeBrujinAst {
    Abstraction(Box<DeBrujinAst>),                   // \:1.2
    Application(Box<DeBrujinAst>, Box<DeBrujinAst>), // 0 1
    FreeVariable(String),                            // x
    BoundVariable(usize),                            // 1
    Constant(Constant),                              // true | false | n
}

impl DeBrujinAst {
    pub fn beta_reduce(self) -> DeBrujinAst {
        match self {
            DeBrujinAst::Application(lhs, rhs) => match *lhs {
                DeBrujinAst::Abstraction(ast) => ast.subst_bound(1, *rhs).beta_reduce(),
                lhs => DeBrujinAst::Application(Box::new(lhs), rhs),
            },
            a => a,
        }
    }

    fn subst_bound(self, depth: usize, subst: DeBrujinAst) -> DeBrujinAst {
        match self {
            DeBrujinAst::Abstraction(ast) => ast.subst_bound(depth + 1, subst),
            DeBrujinAst::Application(lhs, rhs) => DeBrujinAst::Application(
                Box::new(lhs.subst_bound(depth, subst.clone())),
                Box::new(rhs.subst_bound(depth, subst)),
            ),
            DeBrujinAst::BoundVariable(n) if n == depth => subst,
            a => a,
        }
    }
}
