use std::{collections::HashMap, error::Error, rc::Rc};

use crate::{Ast, Constant, Ident, PrimitiveType, Type};

#[derive(Debug)]
pub enum InferError {
    NotAFunction,
    MismatchedType,
    NotInContext,
}

#[cfg(test)]
mod test;

pub fn infer_type(mut gamma: Rc<HashMap<Ident, Type>>, ast: Ast) -> Result<Type, InferError> {
    match ast {
        Ast::Abstraction(arg, arg_type, ast) => {
            Rc::make_mut(&mut gamma).insert(arg, arg_type.clone());
            let out_type = infer_type(gamma, *ast)?;
            Ok(Type::Arrow(Box::new(arg_type), Box::new(out_type)))
        }
        Ast::Application(left, right) => {
            let left_type = infer_type(gamma.clone(), *left)?;
            let Type::Arrow(in_type, out_type) = left_type else {
                return Err(InferError::NotAFunction);
            };
            let right_type = infer_type(gamma, *right)?;
            if *in_type != right_type {
                return Err(InferError::MismatchedType);
            }
            Ok(*out_type)
        }
        Ast::Variable(var) => gamma.get(&var).cloned().ok_or(InferError::NotInContext),
        Ast::Constant(Constant::Nat(_)) => Ok(Type::Primitive(PrimitiveType::Nat)),
        Ast::Constant(Constant::Bool(_)) => Ok(Type::Primitive(PrimitiveType::Bool)),
    }
}
