use sexpr::{Sexpr, parse_string};

use crate::{Ast, Constant, PrimitiveType, Type};

pub mod sexpr;
#[cfg(test)]
mod test;

#[derive(Debug)]
pub enum ParseError {
    UnexpectedParenClose,
    UnexpectedEof,
    TrailingTokens,
    TrailingExpr,
    ToplevelSymbol,
    InvalidSymbol,
    UnexpectedEndOfList,
    UnknownType,
    ExpectedList,
    ExpectedSymbol,
    ExpectedLambda,
    ExpectedIdent,
    ExpectedType,
    NotAType,
    ExpectedBody,
    ExpectedArrow,
    ExpectedOneOf(Vec<String>, String),
}

fn expect_symbol(ast: Option<Sexpr>) -> Result<String, ParseError> {
    match ast {
        Some(Sexpr::Symbol(s)) => Ok(s),
        Some(l) => Err(ParseError::ExpectedSymbol),
        None => Err(ParseError::ExpectedSymbol),
    }
}

fn expect_ident(ast: Option<Sexpr>) -> Result<String, ParseError> {
    let sym = expect_symbol(ast)?;
    if is_ident(&sym) {
        Ok(sym)
    } else {
        Err(ParseError::ExpectedIdent)
    }
}

fn expect_list(ast: Option<Sexpr>) -> Result<Vec<Sexpr>, ParseError> {
    match ast {
        Some(Sexpr::List(l)) => Ok(l),
        Some(l) => Err(ParseError::ExpectedList),
        None => Err(ParseError::ExpectedList),
    }
}

fn expect_one_of<T>(options: &[T], item: String) -> Result<String, ParseError>
where
    T: PartialEq<String> + Into<String> + Clone,
{
    if options.iter().find(|e| **e == item).is_some() {
        Ok(item)
    } else {
        Err(ParseError::ExpectedOneOf(
            options
                .iter()
                .map(|t| Into::<String>::into(t.clone()))
                .collect(),
            item,
        ))
    }
}

fn expect_empty<T, I: Iterator<Item = T>>(mut iter: I) -> Result<(), ParseError> {
    match iter.next() {
        Some(_) => Err(ParseError::TrailingTokens),
        None => Ok(()),
    }
}

pub fn parse(input: &str) -> Result<Ast, ParseError> {
    let ast = parse_string(input)?;
    match ast {
        Sexpr::Symbol(s) => parse_symbol(s),
        list => parse_intern(list),
    }
}

fn parse_intern(ast: Sexpr) -> Result<Ast, ParseError> {
    match ast {
        Sexpr::Symbol(s) => parse_symbol(s),
        Sexpr::List(sexprs) => {
            let mut iter = sexprs.into_iter();
            match iter.next() {
                Some(Sexpr::Symbol(sym)) => {
                    if sym == "\\" {
                        let bind = expect_list(iter.next())?;
                        let mut bind = bind.into_iter();
                        let ident = expect_ident(bind.next())?;
                        let typ = parse_type(&bind.next().ok_or(ParseError::ExpectedType)?)?;
                        expect_empty(bind)?;
                        let ast = Ast::Abstraction(
                            ident,
                            typ,
                            Box::new(parse_intern(iter.next().ok_or(ParseError::ExpectedBody)?)?),
                        );
                        expect_empty(iter)?;
                        Ok(ast)
                    } else {
                        let ast = parse_symbol(sym)?;
                        if let Some(e) = iter.next() {
                            let rhs = parse_intern(e)?;
                            expect_empty(iter)?;
                            Ok(Ast::Application(Box::new(ast), Box::new(rhs)))
                        } else {
                            Ok(ast)
                        }
                    }
                }
                Some(app_left) => {
                    if let Some(app_right) = iter.next() {
                        expect_empty(iter)?;
                        Ok(Ast::Application(
                            Box::new(parse_intern(app_left)?),
                            // Make it back into an Sexpr so we can feed it to parse intern
                            Box::new(parse_intern(app_right)?),
                        ))
                    } else {
                        Err(ParseError::UnexpectedEndOfList)
                    }
                }
                None => Err(ParseError::UnexpectedEndOfList),
            }
        }
    }
}

fn parse_symbol(s: String) -> Result<Ast, ParseError> {
    if let Ok(n) = s.parse::<usize>() {
        Ok(Ast::Constant(Constant::Nat(n)))
    } else if let Ok(b) = s.parse::<bool>() {
        Ok(Ast::Constant(Constant::Bool(b)))
    } else if is_ident(&s) {
        Ok(Ast::Variable(s))
    } else {
        Err(ParseError::InvalidSymbol)
    }
}

pub fn is_ident(s: &str) -> bool {
    s.starts_with(|c: char| c.is_alphabetic()) && s.chars().all(|c| c.is_alphanumeric())
}

pub fn parse_type(ast: &Sexpr) -> Result<Type, ParseError> {
    match ast {
        Sexpr::Symbol(s) => parse_prim_type(s),
        Sexpr::List(sexprs) => parse_type_list(sexprs),
    }
}

fn parse_type_list(typ: &[Sexpr]) -> Result<Type, ParseError> {
    let Some(t) = typ.get(0) else { todo!() };

    if typ.get(1).is_some() {
        let arr = expect_symbol(typ.get(1).cloned())?;
        if arr != "->" {
            return Err(ParseError::ExpectedArrow);
        }
        Ok(Type::Arrow(
            Box::new(parse_type(t)?),
            Box::new(parse_type_list(&typ[2..])?),
        ))
    } else {
        parse_type(t)
    }
}

fn parse_prim_type(typ: &str) -> Result<Type, ParseError> {
    match typ {
        "Bool" => Ok(Type::Primitive(PrimitiveType::Bool)),
        "Nat" => Ok(Type::Primitive(PrimitiveType::Nat)),
        _ => Err(ParseError::UnknownType),
    }
}
